import { Dialog } from "./Dialog";
import { RecordBuffer } from "./RecordBuffer";
import { NullRecord } from "./NullRecord";
import { DataUrl } from "../util/DataUrl";
import { EncodedBinary } from "./EncodedBinary";
import { RedirectionUtil } from "./RedirectionUtil";
import { Property } from "./Property";
import { Attachment } from "./Attachment";
import { TypeNames } from "./types";
import { ViewModeEnum } from "./types";
/**
 * PanContext Subtype that represents an 'Editor Dialog'.
 * An 'Editor' represents and is backed by a single Record and Record definition.
 * See {@link Record} and {@link RecordDef}.
 */
export class EditorDialog extends Dialog {
    /**
     * Get the current buffered record
     * @returns {RecordBuffer}
     */
    get buffer() {
        if (!this._buffer) {
            this._buffer = new RecordBuffer(NullRecord.singleton);
        }
        return this._buffer;
    }
    changeViewMode(viewMode) {
        if (this.viewMode !== viewMode) {
            return this.catavolt.dialogApi.changeMode(this.tenantId, this.sessionId, this.id, viewMode)
                .then((dialog) => {
                //any new dialog needs to be initialized with the Catavolt object
                dialog.initialize(this.catavolt);
                this.updateSettingsWithNewDialogProperties(dialog.referringObject);
                return dialog;
            });
        }
    }
    /**
     * Get the associated entity record
     * @returns {Record}
     */
    get record() {
        return this._buffer.toRecord();
    }
    /**
     * Get the current version of the entity record, with any pending changes present
     * @returns {Record}
     */
    get recordNow() {
        return this.record;
    }
    getAvailableValues(propName) {
        return this.catavolt.dialogApi.getAvailableValues(this.tenantId, this.sessionId, this.id, propName);
    }
    /**
     * Returns whether or not this cell definition contains a binary value
     *
     * @param {AttributeCellValue} cellValue
     * @returns {boolean}
     */
    isBinary(cellValue) {
        const propDef = this.propDefAtName(cellValue.propertyName);
        return propDef && (propDef.isBinaryType || (propDef.isURLType && cellValue.isInlineMediaStyle));
    }
    /**
     * Returns whether or not the buffers contain valid data via a successful read operation.
     * @returns {boolean}
     */
    get isFirstReadComplete() {
        return this._isFirstReadComplete;
    }
    /**
     * Returns whether or not this Editor is in 'read' mode
     * @returns {boolean}
     */
    get isReadMode() {
        return this.viewMode === ViewModeEnum.READ;
    }
    /**
     * Returns whether or not this property is read-only
     * @param propName
     * @returns {boolean}
     */
    isReadModeFor(propName) {
        if (!this.isReadMode) {
            const propDef = this.propDefAtName(propName);
            return !propDef || !propDef.writeAllowed || !propDef.writeEnabled;
        }
        return true;
    }
    /**
     * Returns whether or not this cell definition contains a binary value that should be treated as a signature control
     * @param cellValueDef
     * @returns {PropertyDef|boolean}
     */
    isSignature(cellValueDef) {
        const propDef = this.propDefAtName(cellValueDef.propertyName);
        return this.isBinary(cellValueDef) && propDef.isSignatureType;
    }
    /**
     * Returns whether or not this property is 'writable'
     * @returns {boolean}
     */
    get isWriteMode() {
        return this.viewMode === ViewModeEnum.WRITE;
    }
    performMenuActionWithId(actionId, pendingWrites) {
        return this.invokeMenuActionWithId(actionId, { pendingWrites, type: TypeNames.ActionParametersTypeName })
            .then((result) => {
            return result;
        });
    }
    /**
     * Perform the action associated with the given Menu on this EditorDialog
     * Given that the Editor could possibly be destroyed as a result of this action,
     * any provided pending writes will be saved if present.
     * @param {Menu} menu
     * @param {Record} pendingWrites
     * @returns {Promise<{actionId: string} | Redirection>}
     */
    performMenuAction(menu, pendingWrites) {
        return this.invokeMenuAction(menu, { pendingWrites, type: TypeNames.ActionParametersTypeName })
            .then((result) => {
            return result;
        });
    }
    /**
     * Properties whose {@link PropDef.canCauseSideEffects} value is true, may change other underlying values in the model.
     * This method will update those underlying values, given the property name that is changing, and the new value.
     * This is frequently used with {@link EditorContext.getAvailableValues}.  When a value is seleted, other properties
     * available values may change. (i.e. Country, State, City dropdowns)
     * @param propertyName
     * @param value
     * @returns {Future<null>}
     */
    //@TODO
    processSideEffects(propertyName, value) {
        /*
        var sideEffectsFr: Future<Record> = DialogService.processSideEffects(this.paneDef.dialogHandle,
            this.sessionContext, propertyName, value, this.buffer.afterEffects()).map((changeResult: XPropertyChangeResult) => {
            return changeResult.sideEffects ? changeResult.sideEffects.record : new NullRecord();
        });

        return sideEffectsFr.map((sideEffectsRec: Record) => {
            var originalProps = this.buffer.before.props;
            var userEffects = this.buffer.afterEffects().props;
            var sideEffects = sideEffectsRec.props;
            sideEffects = sideEffects.filter((prop: Prop) => {
                return prop.name !== propertyName;
            });
            this._buffer = RecordBuffer.createRecordBuffer(this.buffer.objectId,
                RecordUtil.union(originalProps, sideEffects),
                RecordUtil.union(originalProps, RecordUtil.union(userEffects, sideEffects)));
            return null;
        });
        */
        return Promise.resolve(null);
    }
    /**
     * Read (load) the {@link Record} assocated with this Editor
     * The record must be read at least once to initialize the Context
     * @returns {Future<Record>}
     */
    read() {
        return this.catavolt.dialogApi.getRecord(this.tenantId, this.sessionId, this.id)
            .then((record) => {
            this._isFirstReadComplete = true;
            this.initBuffer(record);
            this.lastRefreshTime = new Date();
            return this.record;
        });
    }
    /**
     * Set the value of a property in this {@link Record}.
     * Values may be already constructed target types (CodeRef, TimeValue, Date, etc.)
     * or primitives, in which case the values will be parsed and objects constructed as necessary.
     * @param name
     * @param value
     * @returns {any}
     */
    setPropValue(name, value) {
        const propDef = this.propDefAtName(name);
        let parsedValue = null;
        if (propDef) {
            parsedValue = (value !== null && value !== undefined) ? this.parseValue(value, propDef.propertyName) : null;
            this.buffer.setValue(propDef.propertyName, parsedValue);
        }
        return parsedValue;
    }
    /**
     * Set a binary property from a string formatted as a 'data url'
     * See {@link https://en.wikipedia.org/wiki/Data_URI_scheme}
     * @param name
     * @param dataUrl
     */
    setBinaryPropWithDataUrl(name, dataUrl) {
        if (dataUrl) {
            const urlObj = new DataUrl(dataUrl);
            this.setBinaryPropWithEncodedData(name, urlObj.data, urlObj.mimeType);
        }
        else {
            this.setPropValue(name, null); // Property is being deleted/cleared
        }
    }
    /**
     * Set a binary property with base64 encoded data
     * @param name
     * @param encodedData
     * @param mimeType
     */
    setBinaryPropWithEncodedData(name, encodedData, mimeType) {
        const propDef = this.propDefAtName(name);
        if (propDef) {
            const value = new EncodedBinary(encodedData, mimeType);
            this.buffer.setValue(propDef.propertyName, value);
        }
    }
    /**
     * Write this record (i.e. {@link Record}} back to the server
     * @returns {Promise<Record | Redirection>}
     */
    write() {
        let deltaRec = this.buffer.afterEffects();
        /* Write the 'special' props first */
        return this.writeBinaries(deltaRec).then((binResult) => {
            return this.writeAttachments(deltaRec).then((atResult) => {
                /* Remove special property types before writing the actual record */
                deltaRec = this.removeSpecialProps(deltaRec);
                return this.catavolt.dialogApi.putRecord(this.tenantId, this.sessionId, this.id, deltaRec)
                    .then((result) => {
                    const now = new Date();
                    this.catavolt.dataLastChangedTime = now;
                    this.lastRefreshTime = now;
                    if (RedirectionUtil.isRedirection(result)) {
                        this.updateSettingsWithNewDialogProperties(result.referringObject);
                    }
                    else {
                        this.initBuffer(result);
                    }
                    return result;
                });
            });
        });
    }
    //Module level methods
    //Private methods
    /*
        @TODO
        Consider clone and deep copy here, to avoid potential ui side-effects
     */
    removeSpecialProps(record) {
        record.properties = record.properties.filter((prop) => {
            /* Remove the Binary(s) as they have been written seperately */
            return !this.propDefAtName(prop.name).isBinaryType;
        }).map((prop) => {
            /*
             Remove the Attachment(s) (as they have been written seperately) but replace
             the property value with the file name of the attachment prior to writing
             */
            if (prop.value instanceof Attachment) {
                const attachment = prop.value;
                return new Property(prop.name, attachment.name, prop.propertyType, prop.format, prop.annotations);
            }
            else {
                return prop;
            }
        });
        return record;
    }
    initBuffer(record) {
        this._buffer = record ? new RecordBuffer(record) : new RecordBuffer(NullRecord.singleton);
    }
}
