import { Binary } from "./Binary";
import { CatavoltApi } from "../dialog/CatavoltApi";
import { ViewDescriptor } from "./ViewDescriptor";
import { View } from "./View";
import { RecordDef } from "./RecordDef";
import { ReferringObject } from "./ReferringObject";
import { Record } from "./Record";
import { Property } from "./Property";
import { DialogException } from "./DialogException";
import { Menu } from "./Menu";
import { PropertyDef } from "./PropertyDef";
import { Attachment } from "./Attachment";
import { ActionParameters } from "./ActionParameters";
import { Redirection } from "./Redirection";
import { ViewMode } from "./types";
import { DialogType } from "./types";
import { DialogMode } from "./types";
/**
 * Top-level class, representing a Catavolt 'Dialog' definition.
 * All Dialogs have a composite {@link View} definition along with a single record
 * or a list of records.  See {@Record}
 */
export declare abstract class Dialog {
    static BINARY_CHUNK_SIZE: number;
    private static CHAR_CHUNK_SIZE;
    private _binaryCache;
    private _lastRefreshTime;
    private _catavolt;
    readonly availableViews: ViewDescriptor[];
    readonly businessClassName: string;
    readonly children: Dialog[];
    readonly description: string;
    readonly dialogClassName: string;
    dialogMode: DialogMode;
    readonly header: View;
    readonly id: string;
    readonly recordDef: RecordDef;
    readonly referringObject: ReferringObject;
    readonly selectedViewId: string;
    readonly sessionId: string;
    readonly tenantId: string;
    readonly type: DialogType;
    readonly view: View;
    readonly viewMode: ViewMode;
    readonly catavolt: CatavoltApi;
    /**
     * Load a Binary property from a record
     * @param propName
     * @param record
     * @returns {}
     */
    binaryAt(propName: string, record: Record): Promise<Binary>;
    destroy(): void;
    /**
     * Return the error associated with this dialog, if any
     * @returns {}
     */
    readonly error: DialogException;
    /**
     * Find a menu def on this dialog with the given actionId
     * @param actionId
     * @returns {Menu}
     */
    findMenuAt(actionId: string): Menu;
    /**
     * Get a string representation of this property suitable for 'reading'
     *
     * @param {Property} prop
     * @param {string} propName
     * @returns {string}
     */
    formatForRead(prop: Property, propName: string): string;
    /**
     * Get a string representation of this property suitable for 'writing'
     *
     * @param {Property} prop
     * @param {string} propName
     * @returns {string}
     */
    formatForWrite(prop: Property, propName: string): string;
    /**
     * Returns whether or not this dialog loaded properly
     * @returns {boolean}
     */
    readonly hasError: boolean;
    /**
     * Returns whether or not this Form is destroyed
     * @returns {boolean}
     */
    readonly isDestroyed: boolean;
    /**
     * Returns whether or not the data in this dialog is out of date
     * @returns {boolean}
     */
    readonly isRefreshNeeded: boolean;
    /**
     * Get the last time this dialog's data was refreshed
     * @returns {Date}
     */
    /**
     * @param time
     */
    lastRefreshTime: Date;
    /**
     * Get the all {@link Menu}'s associated with this dialog
     * @returns {Array<Menu>}
     */
    readonly menu: Menu;
    openViewWithId(viewId: string): Promise<Dialog>;
    openView(targetViewDescriptor: ViewDescriptor): Promise<Dialog>;
    /**
     * Get the title of this dialog
     * @returns {string}
     */
    readonly paneTitle: string;
    /**
     * Parses a value to prepare for 'writing' back to the server
     * @param formattedValue
     * @param propName
     * @returns {}
     */
    parseValue(formattedValue: any, propName: string): any;
    /**
     * Get the property definition for a property name
     * @param propName
     * @returns {PropertyDef}
     */
    propDefAtName(propName: string): PropertyDef;
    /**
     * Read all the Binary values in this {@link Record}
     *
     * @param {Record} record
     * @returns {Promise<Binary[]>}
     */
    readBinaries(record: Record): Promise<Binary[]>;
    /**
     * Get the all {@link ViewDescriptor}'s associated with this Form
     * @returns {Array<ViewDescriptor>}
     */
    readonly viewDescs: ViewDescriptor[];
    writeAttachment(attachment: Attachment): Promise<void>;
    writeAttachments(record: Record): Promise<void[]>;
    /**
     * Write all Binary values in this {@link Record} back to the server
     *
     * @param {Record} record
     * @returns {Promise<void[]>}
     */
    writeBinaries(record: Record): Promise<void[]>;
    initialize(catavolt: CatavoltApi): void;
    protected invokeMenuActionWithId(actionId: string, actionParams: ActionParameters): Promise<{
        actionId: string;
    } | Redirection>;
    /**
     * Perform this action associated with the given Menu on this dialog.
     * The targets array is expected to be an array of object ids.
     * @param {Menu} menu
     * @param {ActionParameters} actionParams
     * @returns {Promise<{actionId: string} | Redirection>}
     */
    protected invokeMenuAction(menu: Menu, actionParams: ActionParameters): Promise<{
        actionId: string;
    } | Redirection>;
    /**
     *
     * @param {string} propName
     * @param {Record} record
     * @returns {Promise<Binary>}
     */
    protected readBinary(propName: string, record: Record): Promise<Binary>;
    protected updateSettingsWithNewDialogProperties(referringObject: ReferringObject): void;
    /**
     * @private
     * @returns {boolean}
     */
    private readonly isAnyChildDestroyed;
}
